@@ -13,76 +13,96 @@ use crate::Tree;
 ///
 /// This can be used as a checkpoint in [TreeBuilder::close_at], and a
 /// checkpoint can be fetched up front from [TreeBuilder::checkpoint].
-#[derive(Debug, Clone, Copy)]
+#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash, PartialOrd, Ord)]
 #[repr(transparent)]
 pub struct Id(NonMaxUsize);
 
-/// Error raised by [TreeBuilder::close] if there currently is no node being
-/// built.
-///
-/// # Examples
-///
-/// ```
-/// use syntree::{CloseError, Span, TreeBuilder};
-///
-/// # fn main() -> Result<(), Box<dyn std::error::Error>> {
-/// let mut tree = TreeBuilder::new();
-///
-/// tree.open("root");
-/// tree.close()?;
-///
-/// // Syntax::Root and Syntax::Child is left open.
-/// assert!(matches!(tree.close(), Err(CloseError { .. })));
-/// # Ok(()) }
-/// ```
-#[derive(Debug)]
+/// Errors raised while building a tree.
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
 #[non_exhaustive]
-pub struct CloseError;
-
-impl Error for CloseError {}
-
-impl fmt::Display for CloseError {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        write!(f, "no node being built")
-    }
+pub enum TreeBuilderError {
+    /// Error raised by [TreeBuilder::close] if there currently is no node being
+    /// built.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use syntree::{TreeBuilder, TreeBuilderError};
+    ///
+    /// # fn main() -> Result<(), Box<dyn std::error::Error>> {
+    /// let mut tree = TreeBuilder::new();
+    ///
+    /// tree.open("root");
+    /// tree.close()?;
+    ///
+    /// // Syntax::Root and Syntax::Child is left open.
+    /// assert_eq!(tree.close(), Err(TreeBuilderError::CloseError));
+    /// # Ok(()) }
+    /// ```
+    CloseError,
+    /// Error raised by [TreeBuilder::build] if the tree isn't correctly
+    /// balanced.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use syntree::{TreeBuilder, TreeBuilderError};
+    ///
+    /// # fn main() -> Result<(), Box<dyn std::error::Error>> {
+    /// let mut tree = TreeBuilder::new();
+    ///
+    /// tree.open("number");
+    /// tree.token("lit", 3);
+    /// tree.close()?;
+    ///
+    /// tree.open("number");
+    ///
+    /// // Syntax::Number is left open.
+    /// assert_eq!(tree.build(), Err(TreeBuilderError::BuildError));
+    /// # Ok(()) }
+    /// ```
+    BuildError,
+    /// Error raised by [TreeBuilder::close_at] if we're not trying to close at
+    /// a sibling node.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use syntree::{TreeBuilder, TreeBuilderError};
+    ///
+    /// # fn main() -> Result<(), Box<dyn std::error::Error>> {
+    /// let mut tree = TreeBuilder::new();
+    ///
+    /// let c = tree.checkpoint();
+    ///
+    /// tree.open("child");
+    /// tree.token("token", 3);
+    ///
+    /// let result = tree.close_at(c, "operation");
+    /// assert_eq!(result, Err(TreeBuilderError::CloseAtError));
+    /// # Ok(()) }
+    /// ```
+    CloseAtError,
 }
 
-/// Error raised by [TreeBuilder::build] if the tree isn't correctly
-/// balanced.
-///
-/// # Examples
-///
-/// ```
-/// use syntree::{BuildError, Span, TreeBuilder};
-///
-/// #[derive(Debug, Clone, Copy)]
-/// enum Syntax {
-///     Number,
-///     Lit,
-/// }
-///
-/// # fn main() -> Result<(), Box<dyn std::error::Error>> {
-/// let mut tree = TreeBuilder::new();
-///
-/// tree.open(Syntax::Number);
-/// tree.token(Syntax::Lit, 3);
-/// tree.close()?;
-///
-/// tree.open(Syntax::Number);
-///
-/// // Syntax::Number is left open.
-/// assert!(matches!(tree.build(), Err(BuildError { .. })));
-/// # Ok(()) }
-/// ```
-#[derive(Debug)]
-#[non_exhaustive]
-pub struct BuildError;
+impl Error for TreeBuilderError {}
 
-impl Error for BuildError {}
-
-impl fmt::Display for BuildError {
+impl fmt::Display for TreeBuilderError {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        write!(f, "tree is currently being built")
+        match self {
+            TreeBuilderError::CloseError => {
+                write!(f, "no node being built")
+            }
+            TreeBuilderError::BuildError => {
+                write!(f, "tree is currently being built")
+            }
+            TreeBuilderError::CloseAtError => {
+                write!(
+                    f,
+                    "trying to close a node which is not a sibling of the checkpoint being closed"
+                )
+            }
+        }
     }
 }
 
@@ -96,35 +116,29 @@ impl fmt::Display for BuildError {
 /// ```
 /// use syntree::TreeBuilder;
 ///
-/// #[derive(Debug, Clone, Copy, PartialEq, Eq)]
-/// enum Syntax {
-///     Root,
-///     Child,
-/// }
-///
 /// # fn main() -> Result<(), Box<dyn std::error::Error>> {
 /// let mut tree = TreeBuilder::new();
 ///
-/// tree.open(Syntax::Root);
-/// tree.open(Syntax::Child);
+/// tree.open("root");
+/// tree.open("child");
 /// tree.close()?;
-/// tree.open(Syntax::Child);
+/// tree.open("child");
 /// tree.close()?;
 /// tree.close()?;
 ///
 /// let tree = tree.build()?;
 ///
 /// let expected = syntree::tree! {
-///     Syntax::Root => {
-///         Syntax::Child,
-///         Syntax::Child
+///     "root" => {
+///         "child",
+///         "child"
 ///     }
 /// };
 ///
 /// assert_eq!(tree, expected);
 /// # Ok(()) }
 /// ```
-#[derive(Clone)]
+#[derive(Debug, Clone)]
 pub struct TreeBuilder<T> {
     /// Data in the tree being built.
     tree: Tree<T>,
@@ -145,24 +159,32 @@ impl<T> TreeBuilder<T> {
     /// ```
     /// use syntree::TreeBuilder;
     ///
-    /// #[derive(Debug, Clone, Copy)]
-    /// enum Syntax {
-    ///     Root,
-    ///     Child,
-    /// }
-    ///
     /// # fn main() -> Result<(), Box<dyn std::error::Error>> {
     /// let mut tree = TreeBuilder::new();
     ///
-    /// tree.open(Syntax::Root);
+    /// tree.open("root");
     ///
-    /// tree.open(Syntax::Child);
+    /// tree.open("child");
+    /// tree.token("token", 5);
     /// tree.close()?;
     ///
-    /// tree.open(Syntax::Child);
+    /// tree.open("child2");
     /// tree.close()?;
     ///
     /// tree.close()?;
+    ///
+    /// let tree = tree.build()?;
+    ///
+    /// let expected = syntree::tree! {
+    ///     "root" => {
+    ///         "child" => {
+    ///             ("token", 5)
+    ///         },
+    ///         "child2"
+    ///     }
+    /// };
+    ///
+    /// assert_eq!(tree, expected);
     /// # Ok(()) }
     /// ```
     pub const fn new() -> Self {
@@ -185,21 +207,15 @@ impl<T> TreeBuilder<T> {
     /// ```
     /// use syntree::TreeBuilder;
     ///
-    /// #[derive(Debug, Clone, Copy)]
-    /// enum Syntax {
-    ///     Root,
-    ///     Child,
-    /// }
-    ///
     /// # fn main() -> Result<(), Box<dyn std::error::Error>> {
     /// let mut tree = TreeBuilder::new();
     ///
-    /// tree.open(Syntax::Root);
+    /// tree.open("root");
     ///
-    /// tree.open(Syntax::Child);
+    /// tree.open("child");
     /// tree.close()?;
     ///
-    /// tree.open(Syntax::Child);
+    /// tree.open("child");
     /// tree.close()?;
     ///
     /// tree.close()?;
@@ -211,9 +227,10 @@ impl<T> TreeBuilder<T> {
         Id(id)
     }
 
-    /// End a node being built. This call must be balanced with a prior call to
-    /// [TreeBuilder::open] and if its not will result in an
-    /// [CloseError].
+    /// End a node being built.
+    ///
+    /// This call must be balanced with a prior call to [TreeBuilder::open]. If
+    /// not this will result in an [TreeBuilderError::CloseError] being raised.
     ///
     /// This will pop a value of the stack, and set that value as the next
     /// sibling which will be used with [TreeBuilder::open].
@@ -223,30 +240,24 @@ impl<T> TreeBuilder<T> {
     /// ```
     /// use syntree::TreeBuilder;
     ///
-    /// #[derive(Debug, Clone, Copy)]
-    /// enum Syntax {
-    ///     Root,
-    ///     Child,
-    /// }
-    ///
     /// # fn main() -> Result<(), Box<dyn std::error::Error>> {
     /// let mut tree = TreeBuilder::new();
     ///
-    /// tree.open(Syntax::Root);
+    /// tree.open("root");
     ///
-    /// tree.open(Syntax::Child);
+    /// tree.open("child");
     /// tree.close()?;
     ///
-    /// tree.open(Syntax::Child);
+    /// tree.open("child");
     /// tree.close()?;
     ///
     /// tree.close()?;
     /// # Ok(()) }
     /// ```
-    pub fn close(&mut self) -> Result<(), CloseError> {
+    pub fn close(&mut self) -> Result<(), TreeBuilderError> {
         let head = match self.parents.pop() {
             Some(head) => head,
-            None => return Err(CloseError),
+            None => return Err(TreeBuilderError::CloseError),
         };
 
         self.sibling = Some(head);
@@ -273,18 +284,11 @@ impl<T> TreeBuilder<T> {
     /// ```
     /// use syntree::TreeBuilder;
     ///
-    /// #[derive(Debug, Clone, Copy)]
-    /// enum Syntax {
-    ///     Root,
-    ///     Child,
-    ///     Number,
-    /// }
-    ///
     /// # fn main() -> Result<(), Box<dyn std::error::Error>> {
     /// let mut tree = TreeBuilder::new();
     ///
-    /// tree.open(Syntax::Child);
-    /// tree.token(Syntax::Number, 3);
+    /// tree.open("child");
+    /// tree.token("lit", 3);
     /// tree.close()?;
     ///
     /// # Ok(()) }
@@ -309,39 +313,29 @@ impl<T> TreeBuilder<T> {
     /// ```
     /// use syntree::TreeBuilder;
     ///
-    /// #[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
-    /// enum Syntax {
-    ///     ROOT,
-    ///     NUMBER,
-    ///     LIT,
-    ///     WHITESPACE,
-    /// }
-    ///
-    /// use Syntax::*;
-    ///
     /// # fn main() -> Result<(), Box<dyn std::error::Error>> {
     /// let mut b = TreeBuilder::new();
     ///
     /// let c = b.checkpoint();
     ///
-    /// b.open(NUMBER);
-    /// b.token(LIT, 1);
+    /// b.open("number");
+    /// b.token("lit", 1);
     /// b.close()?;
     ///
-    /// b.token(WHITESPACE, 3);
+    /// b.token("whitespace", 3);
     ///
-    /// b.open(NUMBER);
-    /// b.token(LIT, 2);
-    /// b.token(LIT, 2);
+    /// b.open("number");
+    /// b.token("lit", 2);
+    /// b.token("lit", 2);
     /// b.close()?;
     ///
-    /// b.close_at(c, ROOT);
+    /// b.close_at(c, "root")?;
     ///
     /// let tree = b.build()?;
     ///
     /// let root = tree.first().ok_or("missing root")?;
     ///
-    /// assert_eq!(*root.value(), ROOT);
+    /// assert_eq!(*root.value(), "root");
     /// assert_eq!(root.children().count(), 3);
     /// assert_eq!(root.children().without_tokens().count(), 2);
     /// # Ok(()) }
@@ -352,80 +346,65 @@ impl<T> TreeBuilder<T> {
 
     /// Insert a node that wraps from the given checkpointed location.
     ///
-    /// This causes the node specified at `id` to become the previous sibling
-    /// node. If `id` refers to a node that hasn't been allocated yet (through
-    /// [TreeBuilder::checkpoint]), this call corresponds exactly to
-    /// [TreeBuilder::open] that is immediately closed with
-    /// [TreeBuilder::close].
+    /// The checkpoint being closed *must* be a sibling. Otherwise a
+    /// [TreeBuilderError::CloseAtError] will be raised.
+    ///
+    /// # Examples
     ///
     /// ```
     /// use syntree::TreeBuilder;
     ///
     /// # fn main() -> Result<(), Box<dyn std::error::Error>> {
-    /// let mut a = TreeBuilder::<u32>::new();
-    /// a.open(0);
-    /// let c = a.checkpoint();
-    /// a.close_at(c, 1);
-    /// a.close()?;
-    /// let a = a.build()?;
+    /// let mut b = TreeBuilder::new();
     ///
-    /// let mut b = TreeBuilder::<u32>::new();
-    /// b.open(0);
     /// let c = b.checkpoint();
-    /// b.close_at(c, 1);
-    /// b.close()?;
-    /// let b = b.build()?;
+    /// b.token("lit", 3);
+    /// b.close_at(c, "root")?;
     ///
-    /// assert_eq!(a, b);
+    /// let tree = b.build()?;
+    ///
+    /// let expected = syntree::tree! {
+    ///     "root" => {
+    ///         ("lit", 3)
+    ///     }
+    /// };
+    ///
+    /// assert_eq!(tree, expected);
     /// # Ok(()) }
     /// ```
     ///
-    /// Note that this does not modify or try to balance the stack, so the last
-    /// item pushed using [TreeBuilder::open] will still be the one popped
-    /// through [TreeBuilder::close].
-    ///
-    /// # Examples
+    /// More complex example:
     ///
     /// ```
-    /// use syntree::{print, Span, TreeBuilder};
-    ///
-    /// #[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
-    /// enum Syntax {
-    ///     ROOT,
-    ///     NUMBER,
-    ///     LIT,
-    ///     WHITESPACE,
-    /// }
-    ///
-    /// use Syntax::*;
+    /// use syntree::TreeBuilder;
     ///
     /// # fn main() -> Result<(), Box<dyn std::error::Error>> {
     /// let mut b = TreeBuilder::new();
     ///
     /// let c = b.checkpoint();
     ///
-    /// b.open(NUMBER);
-    /// b.token(LIT, 3);
+    /// b.open("number");
+    /// b.token("lit", 3);
     /// b.close()?;
     ///
-    /// b.token(WHITESPACE, 1);
+    /// b.token("whitespace", 1);
     ///
-    /// b.open(NUMBER);
-    /// b.token(LIT, 2);
+    /// b.open("number");
+    /// b.token("lit", 2);
     /// b.close()?;
     ///
-    /// b.close_at(c, ROOT);
+    /// b.close_at(c, "root")?;
     ///
     /// let tree = b.build()?;
     ///
     /// let expected = syntree::tree! {
-    ///     ROOT => {
-    ///         NUMBER => {
-    ///             (LIT, 3)
+    ///     "root" => {
+    ///         "number" => {
+    ///             ("lit", 3)
     ///         },
-    ///         (WHITESPACE, 1),
-    ///         NUMBER => {
-    ///             (LIT, 2)
+    ///         ("whitespace", 1),
+    ///         "number" => {
+    ///             ("lit", 2)
     ///         },
     ///     }
     /// };
@@ -433,9 +412,7 @@ impl<T> TreeBuilder<T> {
     /// assert_eq!(tree, expected);
     /// # Ok(()) }
     /// ```
-    pub fn close_at(&mut self, id: Id, data: T) -> Id {
-        // With the layout of this data structure this is a fairly simple
-        // operation.
+    pub fn close_at(&mut self, id: Id, data: T) -> Result<Id, TreeBuilderError> {
         let child = NonMaxUsize::new(self.tree.len()).expect("ran out of ids");
 
         let links = match self.tree.get_mut(id.0) {
@@ -443,10 +420,12 @@ impl<T> TreeBuilder<T> {
             None => {
                 let id = self.insert(data, Kind::Node, Span::point(self.cursor));
                 self.sibling = Some(id);
-                return Id(id);
+                return Ok(Id(id));
             }
         };
 
+        let needs_restructuring = self.sibling != Some(id.0);
+
         let removed = mem::replace(
             links,
             Links {
@@ -458,16 +437,32 @@ impl<T> TreeBuilder<T> {
             },
         );
 
-        // Adjust span to encapsulate all children.
-        let start = links.span.start;
+        if needs_restructuring {
+            // Adjust span to encapsulate all children and check that we just
+            // inserted ourselves in the right location.
+            let start = links.span.start;
 
-        if let Some(node) = self.tree.node_at(removed.next) {
-            if let Some(last) = node.children().last() {
-                let span = Span::new(start, last.span().end);
+            let sibling = self
+                .tree
+                .node_at(self.sibling)
+                .ok_or(TreeBuilderError::CloseAtError)?;
+
+            if let Some(mut node) = self.tree.node_at(removed.next) {
+                while let Some(next) = node.next() {
+                    node = next;
+                }
+
+                let span = Span::new(start, node.span().end);
+
+                if !sibling.is_same(&node) {
+                    return Err(TreeBuilderError::CloseAtError);
+                }
 
                 if let Some(node) = self.tree.get_mut(id.0) {
                     node.span = span;
                 }
+            } else if !sibling.is_same_as_links(&removed) {
+                return Err(TreeBuilderError::CloseAtError);
             }
         }
 
@@ -475,41 +470,35 @@ impl<T> TreeBuilder<T> {
 
         // The current sibling is the newly replaced node in the tree.
         self.sibling = Some(id.0);
-        Id(id.0)
+        Ok(Id(id.0))
     }
 
     /// Build a [Tree] from the current state of the builder.
     ///
     /// This requires the stack in the builder to be empty. Otherwise a
-    /// [BuildError] will be raised.
+    /// [TreeBuilderError::BuildError] will be raised.
     ///
     /// # Examples
     ///
     /// ```
     /// use syntree::TreeBuilder;
     ///
-    /// #[derive(Debug, Clone, Copy, PartialEq, Eq)]
-    /// enum Syntax {
-    ///     Child,
-    ///     Number,
-    /// }
-    ///
     /// # fn main() -> Result<(), Box<dyn std::error::Error>> {
     /// let mut tree = TreeBuilder::new();
     ///
-    /// tree.open(Syntax::Child);
-    /// tree.token(Syntax::Number, 3);
+    /// tree.open("child");
+    /// tree.token("number", 3);
     /// tree.close()?;
-    /// tree.open(Syntax::Child);
+    /// tree.open("child");
     /// tree.close()?;
     ///
     /// let tree = tree.build()?;
     ///
     /// let expected = syntree::tree! {
-    ///     Syntax::Child => {
-    ///         (Syntax::Number, 3)
+    ///     "child" => {
+    ///         ("number", 3)
     ///     },
-    ///     Syntax::Child,
+    ///     "child",
     /// };
     ///
     /// assert_eq!(tree, expected);
@@ -519,30 +508,24 @@ impl<T> TreeBuilder<T> {
     /// If a tree is unbalanced during construction, building will fail with an error:
     ///
     /// ```
-    /// use syntree::{BuildError, Span, TreeBuilder};
-    ///
-    /// #[derive(Debug, Clone, Copy)]
-    /// enum Syntax {
-    ///     Number,
-    ///     Lit,
-    /// }
+    /// use syntree::{TreeBuilderError, Span, TreeBuilder};
     ///
     /// # fn main() -> Result<(), Box<dyn std::error::Error>> {
     /// let mut tree = TreeBuilder::new();
     ///
-    /// tree.open(Syntax::Number);
-    /// tree.token(Syntax::Lit, 3);
+    /// tree.open("number");
+    /// tree.token("lit", 3);
     /// tree.close()?;
     ///
-    /// tree.open(Syntax::Number);
+    /// tree.open("number");
     ///
-    /// // Syntax::Number is left open.
-    /// assert!(matches!(tree.build(), Err(BuildError { .. })));
+    /// // "number" is left open.
+    /// assert!(matches!(tree.build(), Err(TreeBuilderError::BuildError)));
     /// # Ok(()) }
     /// ```
-    pub fn build(self) -> Result<Tree<T>, BuildError> {
+    pub fn build(self) -> Result<Tree<T>, TreeBuilderError> {
         if !self.parents.is_empty() {
-            return Err(BuildError);
+            return Err(TreeBuilderError::BuildError);
         }
 
         Ok(self.tree)
